import asyncio
from typing import Optional, Callable, Any, TypeAlias, Dict, List

from aio_pika.abc import (
    AbstractIncomingMessage,
    AbstractConnection,
    AbstractQueue,
    AbstractChannel,
)

from .exchange_types import RabbitExchangeType


Arguments: TypeAlias = Dict[str, str | int | None]


class Consumer:
    def __init__(
        self,
        callback: Callable[[bytes], Any],
        queue: str,
        exchange: str,
        exchange_type: RabbitExchangeType,
        routing_key: str,
        prefetch_count: Optional[int] = None,
        auto_ack: bool = False,
        arguments: Optional[Arguments] = None,

        queue_durable: bool = False,
        queue_auto_delete: bool = False,
        queue_passive: bool = False,
        queue_exclusive: bool = False,
        queue_arguments: Optional[Arguments] = None,

        exchange_durable: bool = False,
        exchange_auto_delete: bool = False,
        exchange_internal: bool = False,
        exchange_passive: bool = False,
        exchange_arguments: Optional[Arguments] = None,
    ) -> None:

        self.cb = callback
        self.queue_name = queue
        self.exchange_name = exchange
        self.exchange_type = exchange_type.value
        self.routing_key = routing_key
        self.prefetch_count = prefetch_count
        self.auto_ack = auto_ack
        self.arguments = arguments

        self._queue_args = {
            'durable': queue_durable,
            'auto_delete': queue_auto_delete,
            'passive': queue_passive,
            'exclusive': queue_exclusive,
            'arguments': queue_arguments,
        }
        self._exchange_args = {
            'durable': exchange_durable,
            'auto_delete': exchange_auto_delete,
            'internal': exchange_internal,
            'passive': exchange_passive,
            'arguments': exchange_arguments,
        }

        self.consumer_tag: str = ""
        self._queue: AbstractQueue

    async def setup(self, channel: AbstractChannel) -> AbstractQueue:
        if self.prefetch_count:
            await channel.set_qos(self.prefetch_count)

        exchange = await channel.declare_exchange(
            self.exchange_name,
            self.exchange_type,
            **self._exchange_args)  # type: ignore
        queue = await channel.declare_queue(
            self.queue_name,
            **self._queue_args)  # type: ignore
        await queue.bind(exchange, self.routing_key)

        return queue

    async def _wrapped_callback(self, msg: AbstractIncomingMessage) -> Any:
        async with msg.process():
            return await self.cb(msg.body)

    async def start(self, conn: AbstractConnection) -> str:
        channel = await conn.channel()
        self._queue = await self.setup(channel)

        self._running_tasks: set = set()

        async with self._queue.iterator(
            no_ack=self.auto_ack, arguments=self.arguments
        ) as queue_iterator:
            self.consumer_tag = queue_iterator._consumer_tag

            async for msg in queue_iterator:
                task = asyncio.create_task(self._wrapped_callback(msg))
                task.add_done_callback(self._running_tasks.discard)
                self._running_tasks.add(task)

        return self.consumer_tag

    async def stop(
        self, timeout: Optional[int] = None, nowait: bool = False
    ) -> List[Exception | None]:
        await self._queue.cancel(
            self.consumer_tag, timeout=timeout, nowait=nowait)

        return await asyncio.gather(*self._running_tasks, return_exceptions=True)



author_info = (
    ("Mehdi Kamani", "mkamani@sheypoor.com"),)

package_info = """
Asyncio compatible wraper around the aio-pika library for consumining messages.
"""
package_license = "MIT License"

team_email = ""

version_info = (0, 0, 2)

__author__ = ", ".join("{} <{}>".format(*info) for info in author_info)
__version__ = ".".join(map(str, version_info))



import asyncio
import functools
from typing import Callable, Optional, TypeAlias, TypeVar, List, Dict

import aio_pika
from aio_pika.abc import AbstractConnection

from .consumer import Consumer
from .exchange_types import RabbitExchangeType


Arguments: TypeAlias = Dict[str, str | int | None]
T = TypeVar("T")


class Bunny:
    _conn: AbstractConnection

    def __init__(self, amqp_url: str) -> None:
        self.url = amqp_url

        self._consumers: List[Consumer] = []

    async def _connect(self) -> AbstractConnection:
        return await aio_pika.connect_robust(self.url)

    async def start(self) -> None:
        self._conn = await self._connect()

        await asyncio.gather(
            *[handler.start(self._conn) for handler in self._consumers])

    async def stop(self, timeout: int = None, nowait: bool = False) -> None:
        await asyncio.gather(
            *[consumer.stop(timeout=timeout, nowait=nowait)
                for consumer in self._consumers])

    def consumer(
        self,
        queue: str,
        exchange: str,
        exchange_type: RabbitExchangeType = RabbitExchangeType.DIRECT,
        routing_key: Optional[str] = None,
        prefetch_count: Optional[int] = None,
        auto_ack: bool = False,
        consumer_arguments: Optional[Arguments] = None,

        queue_durable: bool = False,
        queue_auto_delete: bool = False,
        queue_passive: bool = False,
        queue_exclusive: bool = False,
        queue_arguments: Optional[Arguments] = None,

        exchange_durable: bool = False,
        exchange_auto_delete: bool = False,
        exchange_internal: bool = False,
        exchange_passive: bool = False,
        exchange_arguments: Optional[Arguments] = None,
    ) -> Callable:
        def decorator(func: Callable[[bytes], T]) -> Callable:
            self._consumers.append(
                Consumer(
                    func,
                    queue,
                    exchange,
                    exchange_type,
                    routing_key or exchange,
                    prefetch_count=prefetch_count,
                    auto_ack=auto_ack,
                    arguments=consumer_arguments,

                    queue_durable=queue_durable,
                    queue_auto_delete=queue_auto_delete,
                    queue_passive=queue_passive,
                    queue_exclusive=queue_exclusive,
                    queue_arguments=queue_arguments,

                    exchange_durable=exchange_durable,
                    exchange_auto_delete=exchange_auto_delete,
                    exchange_internal=exchange_internal,
                    exchange_passive=exchange_passive,
                    exchange_arguments=exchange_arguments))

            @functools.wraps(func)
            def _decorator(msg: bytes) -> T:
                return func(msg)

            return _decorator
        return decorator



import os
from setuptools import setup, find_packages
from importlib.machinery import SourceFileLoader


module = SourceFileLoader(
    "version", os.path.join("aio_bunny", "version.py")
).load_module()


setup(
    name="aio-bunny",
    version=module.__version__,
    author=module.__author__,
    author_email=module.team_email,
    license=module.package_license,
    description=module.package_info,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    platforms="all",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Topic :: Internet",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Microsoft",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=find_packages(exclude=["tests"]),
    package_data={"aio_bunny": ["py.typed"]},
    install_requires=["aio-pika>=6.8.0,<10", ],
    python_requires=">=3.10, <4",
    extras_require={
        "develop": [
            "aio-pika",
            "async_generator",
            "coverage!=4.3",
            "coveralls",
            "pylama",
            "pytest",
            "pytest-asyncio",
            "pytest-cov",
            "shortuuid",
            "nox",
            "sphinx",
            "sphinx-autobuild",
            "timeout-decorator",
            "tox>=2.4",
        ],
    },
    project_urls={
        "Documentation": "https://aio-bunny.readthedocs.org/",
        "Source": "https://github.com/itsmehdi97/aio-bunny",
    },
)
